"""

File input/output for localization data.

"""

from __future__ import annotations

import logging
import os
from collections.abc import Callable
from enum import Enum
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from _typeshed import SupportsRead

import pandas as pd

import locan.constants
from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.locan_io.locdata.asdf_io import load_asdf_file
from locan.locan_io.locdata.decode_io import load_decode_file
from locan.locan_io.locdata.elyra_io import load_Elyra_file
from locan.locan_io.locdata.nanoimager_io import load_Nanoimager_file
from locan.locan_io.locdata.rapidstorm_io import (
    load_rapidSTORM_file,
    load_rapidSTORM_track_file,
)
from locan.locan_io.locdata.smap_io import load_SMAP_file
from locan.locan_io.locdata.smlm_io import load_SMLM_file
from locan.locan_io.locdata.thunderstorm_io import load_thunderstorm_file
from locan.locan_io.locdata.utilities import (
    convert_property_names,
    convert_property_types,
)

__all__: list[str] = ["load_txt_file", "load_locdata"]

logger = logging.getLogger(__name__)


def load_txt_file(
    path: str | os.PathLike[str] | SupportsRead[Any],
    sep: str = ",",
    columns: list[str] | None = None,
    nrows: int | None = None,
    property_mapping: dict[str, str] | list[dict[str, str]] | None = None,
    convert: bool = True,
    **kwargs: Any,
) -> LocData:
    """
    Load localization data from a txt file.

    Locan column names are either supplied or read from the first line header.

    Parameters
    ----------
    path
        File path for a localization file to load.
    sep
        separator between column values (Default: ',')
    columns
        Locan column names. If None the first line is interpreted as header
        (Default: None).
    nrows
        The number of localizations to load from file. None means that all
        available rows are loaded (Default: None).
    property_mapping
        Mappings between column names and locan property names
    convert
        If True convert types by applying type specifications in
        locan.constants.PROPERTY_KEYS.
    kwargs
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    # define columns
    if columns is None:
        dataframe: pd.DataFrame = pd.read_csv(  # type:ignore[assignment]
            path,  # type:ignore[arg-type]
            sep=sep,
            nrows=nrows,
            **dict(dict(skiprows=0), **kwargs),
        )

        column_keys = convert_property_names(
            dataframe.columns, property_mapping=property_mapping
        )
        dataframe.columns = column_keys  # type: ignore[assignment]
    else:
        column_keys = convert_property_names(columns, property_mapping=property_mapping)
        dataframe = pd.read_csv(  # type:ignore[assignment]
            path,  # type:ignore[arg-type]
            sep=sep,
            nrows=nrows,
            **dict(dict(skiprows=1, names=column_keys), **kwargs),
        )

    if convert:
        dataframe = convert_property_types(
            dataframe, types=locan.constants.PROPERTY_KEYS
        )

    dat = LocData.from_dataframe(dataframe=dataframe)

    dat.meta.source = metadata_pb2.EXPERIMENT
    dat.meta.state = metadata_pb2.RAW
    dat.meta.file.type = metadata_pb2.CUSTOM
    dat.meta.file.path = str(path)

    del dat.meta.history[:]
    dat.meta.history.add(
        name="load_txt_file",
        parameter=f"path={path}, sep={sep}, columns={columns}, nrows={nrows}",
    )

    return dat


def _map_file_type_to_load_function(
    file_type: int | str | locan.constants.FileType | locan.data.metadata_pb2.Metadata,
) -> Callable[..., Any]:
    """
    Interpret user input for file_type.

    Parameters
    ----------
    file_type
        Identifier for the file type. Integer or string should be according to
        locan.constants.FileType.

    Returns
    -------
    Callable[..., Any]
        Name of function for loading the localization file of `type`.
    """
    look_up_table = dict(
        load_txt_file=load_txt_file,
        load_rapidSTORM_file=load_rapidSTORM_file,
        load_Elyra_file=load_Elyra_file,
        load_thunderstorm_file=load_thunderstorm_file,
        load_asdf_file=load_asdf_file,
        load_Nanoimager_file=load_Nanoimager_file,
        load_rapidSTORM_track_file=load_rapidSTORM_track_file,
        load_SMLM_file=load_SMLM_file,
        load_decode_file=load_decode_file,
        load_SMAP_file=load_SMAP_file,
    )

    class LoadFunction(Enum):
        load_txt_file = 1
        load_rapidSTORM_file = 2
        load_Elyra_file = 3
        load_thunderstorm_file = 4
        load_asdf_file = 5
        load_Nanoimager_file = 6
        load_rapidSTORM_track_file = 7
        load_SMLM_file = 8
        load_decode_file = 9
        load_SMAP_file = 10

    try:
        if isinstance(file_type, int):
            function_name = LoadFunction(file_type).name
        elif isinstance(file_type, str):
            function_name = LoadFunction(
                locan.constants.FileType[file_type.upper()].value
            ).name
        elif isinstance(file_type, locan.constants.FileType):
            function_name = LoadFunction(file_type.value).name
        elif isinstance(file_type, metadata_pb2):  # type: ignore
            function_name = LoadFunction(file_type).name
        else:
            raise TypeError
        return look_up_table[function_name]  # type: ignore
    except ValueError as exc:
        raise ValueError(f"There is no load function for type {file_type}.") from exc


def load_locdata(
    path: str | os.PathLike[Any] | SupportsRead[Any],
    file_type: (
        int | str | locan.constants.FileType | locan.data.metadata_pb2.Metadata
    ) = 1,
    nrows: int | None = None,
    **kwargs: Any,
) -> LocData:
    """
    Load data from localization file as specified by type.

    This function is a wrapper for read functions for the various types of SMLM data.

    Parameters
    ----------
    path
        File path for a localization data file to load.
    file_type
        Indicator for the file type.
        Integer or string should be according to locan.constants.FileType.
    nrows
        The number of localizations to load from file. None means that all
        available rows are loaded.
    kwargs
        kwargs passed to the specific load function.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    return_value: LocData = _map_file_type_to_load_function(file_type)(
        path=path, nrows=nrows, **kwargs
    )
    return return_value



"""

File input/output for localization data in rapidSTORM files.

"""

from __future__ import annotations

import logging
import os
import xml.etree.ElementTree as etree  # noqa: S405
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from _typeshed import SupportsRead, SupportsReadline

import numpy as np
import pandas as pd

import locan.constants
from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.locan_io.locdata.utilities import (
    convert_property_names,
    convert_property_types,
    open_path_or_file_like,
)

__all__: list[str] = [
    "load_rapidSTORM_header",
    "load_rapidSTORM_file",
    "load_rapidSTORM_track_header",
    "load_rapidSTORM_track_file",
]

logger = logging.getLogger(__name__)


def _read_rapidSTORM_header(file: SupportsReadline[Any]) -> list[str]:
    """
    Read xml header from a rapidSTORM single-molecule localization file and
    identify column names.

    Parameters
    ----------
    file : SupportsReadline
        A rapidSTORM file to load.

    Returns
    -------
    list[str]
        A list of valid dataset property keys as derived from the rapidSTORM
        identifiers.
    """
    # read xml part in header
    header = file.readline()
    header = header[2:]

    # get iteratible
    parsed = etree.XML(header)

    # list identifiers
    identifiers = []
    for elem in parsed:
        for name, value in sorted(elem.attrib.items()):
            if name == "identifier":
                identifiers.append(value)

    # turn identifiers into valuable LocData keys
    column_keys = convert_property_names(
        properties=identifiers, property_mapping=locan.constants.RAPIDSTORM_KEYS
    )
    return column_keys


def load_rapidSTORM_header(
    path: str | os.PathLike[Any] | SupportsRead[Any],
) -> list[str]:
    """
    Load xml header from a rapidSTORM single-molecule localization file and
    identify column names.

    Parameters
    ----------
    path
        File path for a file to load.

    Returns
    -------
    list[str]
        A list of valid dataset property keys as derived from the
        identifiers.
    """

    # read xml part in header
    with open_path_or_file_like(path) as file:
        return _read_rapidSTORM_header(file)


def load_rapidSTORM_file(
    path: str | os.PathLike[Any] | SupportsRead[Any],
    nrows: int | None = None,
    convert: bool = True,
    **kwargs: Any,
) -> LocData:
    """
    Load data from a rapidSTORM single-molecule localization file.

    Parameters
    ----------
    path
        File path for a file to load.
    nrows
        The number of localizations to load from file. None means that all
        available rows are loaded.
    convert
        If True convert types by applying type specifications in
        locan.constants.PROPERTY_KEYS.
    kwargs
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with open_path_or_file_like(path) as file:
        columns = _read_rapidSTORM_header(file)
        dataframe = pd.read_csv(
            file, sep=" ", skiprows=0, nrows=nrows, names=columns, **kwargs
        )

    if convert:
        dataframe = convert_property_types(
            dataframe, types=locan.constants.PROPERTY_KEYS
        )

    dat = LocData.from_dataframe(dataframe=dataframe)

    dat.meta.source = metadata_pb2.EXPERIMENT
    dat.meta.state = metadata_pb2.RAW
    dat.meta.file.type = metadata_pb2.RAPIDSTORM
    dat.meta.file.path = str(path)

    for property in sorted(
        list(set(columns).intersection({"position_x", "position_y", "position_z"}))
    ):
        dat.meta.localization_properties.add(name=property, unit="nm", type="float")

    del dat.meta.history[:]
    dat.meta.history.add(
        name="load_rapidSTORM_file", parameter=f"path={str(path)}, nrows={nrows}"
    )

    return dat


def _read_rapidSTORM_track_header(
    file: SupportsReadline[Any],
) -> tuple[list[str], list[str]]:
    """
    Read xml header from a rapidSTORM (track) single-molecule localization
    file and identify column names.

    Parameters
    ----------
    file
        A file to load.

    Returns
    -------
    tuple[list[str], list[str]]
        A list of valid dataset property keys as derived from the
        identifiers.
    """
    # read xml part in header
    header = file.readline()
    header = header[2:]

    parsed = etree.XML(header)

    # list identifiers
    identifiers: list[str] = []
    for field in parsed.findall("field"):
        next_identifier = field.get("identifier")
        if next_identifier is not None:
            identifiers.append(next_identifier)

    # turn identifiers into valuable LocData keys
    column_keys = convert_property_names(
        properties=identifiers, property_mapping=locan.constants.RAPIDSTORM_KEYS
    )

    # list child identifiers
    child_identifiers: list[str] = []
    for field in parsed.findall("localizations"):
        for field_ in field.findall("field"):
            next_child_identifiers = field_.get("identifier")
            if next_child_identifiers is not None:
                child_identifiers.append(next_child_identifiers)

    # turn child identifiers into valuable LocData keys
    column_keys_tracks = convert_property_names(
        properties=child_identifiers, property_mapping=locan.constants.RAPIDSTORM_KEYS
    )

    return column_keys, column_keys_tracks


def load_rapidSTORM_track_header(
    path: str | os.PathLike[Any] | SupportsRead[Any],
) -> tuple[list[str], list[str]]:
    """
    Load xml header from a rapidSTORM (track) single-molecule localization
    file and identify column names.

    Parameters
    ----------
    path
        File path for a file to load.

    Returns
    -------
    tuple[list[str], list[str]]
        A list of valid dataset property keys as derived from the
        identifiers.
    """

    # read xml part in header
    with open_path_or_file_like(path) as file:
        return _read_rapidSTORM_track_header(file)


def load_rapidSTORM_track_file(
    path: str | os.PathLike[Any] | SupportsRead[Any],
    nrows: int | None = None,
    convert: bool = True,
    collection: bool = True,
    min_localization_count: int = 1,
    **kwargs: Any,
) -> LocData:
    """
    Load data from a rapidSTORM single-molecule localization file with
    tracked localizations.

    Parameters
    ----------
    path
        File path for a file to load.
    nrows
        The number of localizations to load from file. None means that all
        available rows are loaded.
    convert
        If True convert types by applying type specifications in
        locan.constants.PROPERTY_KEYS.
    collection
        If True a collection of all tracks is returned.
        If False LocData with center positions is returned.
    min_localization_count
        If collection is True, only clusters with at least
        `min_localization_count` localizations are loaded.
    kwargs
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations/tracks as a
        collection.
    """
    with open_path_or_file_like(path) as file:
        columns, columns_track = _read_rapidSTORM_track_header(file)
        lines = pd.read_csv(
            file, lineterminator="\n", nrows=nrows, skiprows=1, header=None, **kwargs
        )

    lines = lines[0].str.split(" ", expand=False)

    if collection:
        # prepare dataframes with tracked localizations
        tracks = [
            np.array(line[len(columns) + 1 :]).reshape(-1, len(columns_track))
            for line in lines
            if int(line[len(columns)]) >= min_localization_count
        ]
        # +1 to account for the column with number of locs in track
        track_list = []
        for track in tracks:
            dataframe = pd.DataFrame(track, columns=columns_track)
            if convert:
                dataframe = convert_property_types(
                    dataframe, types=locan.constants.PROPERTY_KEYS
                )
            else:
                dataframe = dataframe.convert_dtypes()
            dat = LocData.from_dataframe(dataframe=dataframe)
            track_list.append(dat)

        new_collection = LocData.from_collection(track_list)

        new_collection.meta.source = metadata_pb2.EXPERIMENT
        new_collection.meta.state = metadata_pb2.RAW
        new_collection.meta.file.type = metadata_pb2.RAPIDSTORMTRACK
        new_collection.meta.file.path = str(path)

        for property in sorted(
            list(
                set(columns_track).intersection(
                    {"position_x", "position_y", "position_z"}
                )
            )
        ):
            new_collection.meta.localization_properties.add(
                name=property, unit="nm", type="float"
            )

        del new_collection.meta.history[:]
        new_collection.meta.history.add(
            name="load_rapidSTORM_track_file",
            parameter=f"path={str(path)}, nrows={nrows}",
        )

        return new_collection

    else:
        # prepare dataframe with center track positions
        dataframe = pd.DataFrame(
            [line[: len(columns)] for line in lines], columns=columns
        )
        if convert:
            dataframe = convert_property_types(
                dataframe, types=locan.constants.PROPERTY_KEYS
            )
        else:
            dataframe = dataframe.convert_dtypes()

        locdata = LocData.from_dataframe(dataframe=dataframe)

        locdata.meta.source = metadata_pb2.EXPERIMENT
        locdata.meta.state = metadata_pb2.RAW
        locdata.meta.file.type = metadata_pb2.RAPIDSTORM
        locdata.meta.file.path = str(path)

        for property in sorted(
            list(set(columns).intersection({"position_x", "position_y", "position_z"}))
        ):
            locdata.meta.localization_properties.add(
                name=property, unit="nm", type="float"
            )

        del locdata.meta.history[:]
        locdata.meta.history.add(
            name="load_rapidSTORM_track_file",
            parameter=f"path={str(path)}, nrows={nrows}",
        )

        return locdata



"""
Building an analysis pipeline.

Pipeline refers to sequential analysis steps that are applied to a single
LocData object.
An analysis pipeline here includes true piped analysis, where a preliminary
result serves as input to the next analysis
step, but also workflows that provide different results in parallel.

A batch process is a procedure for running a pipeline over multiple LocData
objects while collecting and combing
results.

This module provides a class `Pipeline` to combine the analysis procedure,
parameters and results in a single pickleable object.
"""

from __future__ import annotations

import inspect
import logging
import os
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, TypeVar

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis

if TYPE_CHECKING:
    from locan.data.locdata import LocData

__all__: list[str] = ["Pipeline"]

logger = logging.getLogger(__name__)


class Pipeline(_Analysis):
    """
    The base class for a specialized analysis pipeline to be used on LocData
    objects.

    The custom analysis routine has to be added by implementing the method
    `computation(self, **kwargs)`.
    Keyword arguments must include the locdata reference and optional
    parameters.

    Results are provided as customized attributes.
    We suggest abbreviated standard names for the most common procedures
    such as:

    * lp - Localization Precision
    * lprop - Localization Property
    * lpf - Localizations per Frame
    * rhf - Ripley H function
    * clust - locdata with clustered elements

    Parameters
    ----------
    computation : Callable[..., Any]
        A function `computation(self, **kwargs)` specifying the analysis
        procedure.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    kwargs
        Locdata reference and optional parameters passed to
        `computation(self, **kwargs)`.

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata | None
        Metadata about the current analysis routine.
    computation : Callable[..., Any]
        A function `computation(self, **kwargs)` specifying the analysis
        procedure.
    kwargs
        All parameters including the locdata reference that are passed to
        `computation(self, **kwargs)`.

    Note
    ----
    The class variable `Pipeline.count` is only incremented in a single
    process. In multiprocessing `Pipeline.count` and
    `Pipeline.meta.identifier` (which is set using `count`) cannot be used to
    identify distinct Pipeline objects.

    Note
    ----
    For the Pipeline object to be pickleable attention has to be paid to the
    :func:`computation` method.
    With multiprocessing it will have to be re-injected for each Pipeline
    object by `pipeline.computation = computation`
    after computation and before pickling.
    """

    def __init__(
        self,
        computation: Callable[..., Any],
        meta: metadata_analysis_pb2.AMetadata | None = None,
        **kwargs: Any,
    ) -> None:
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)

        if not callable(computation):
            raise TypeError(
                "A callable function `computation(self, locdata, **kwargs)` "
                "must be passed as first argument."
            )
        self.computation = computation
        self.kwargs = kwargs

    def __bool__(self) -> bool:
        return True

    def compute(self) -> Any:
        """
        Run the analysis procedure. All parameters must be given upon Pipeline
        instantiation.
        """
        return self.computation(self, **self.kwargs)

    def save_computation(self, path: str | os.PathLike[Any]) -> None:
        """
        Save the analysis procedure (i.e. the computation() method) as human
        readable text.

        Parameters
        ----------
        path : str | os.PathLike[Any]
            Path and file name for saving the text file.
        """
        with open(path, "w") as handle:
            handle.write(f"Analysis Pipeline: {self.__class__.__name__}\n\n")
            handle.write(inspect.getsource(self.computation))

    def computation_as_string(self) -> str:
        """
        Return the analysis procedure (i.e. the computation() method) as string.
        """
        return inspect.getsource(self.computation)


T_Pipeline = TypeVar("T_Pipeline", bound="Pipeline")


def computation_test(
    self: T_Pipeline,
    locdata: LocData | None = None,
    parameter: str = "test",
) -> T_Pipeline:
    """A pipeline definition for testing."""
    self.locdata = locdata  # type: ignore
    something = "changed_value"
    logger.debug(f"something has a : {something}")
    self.test = parameter  # type: ignore
    logger.info(f"computation finished for locdata: {locdata}")

    try:
        raise NotImplementedError
    except NotImplementedError:
        logger.warning(f"An exception occurred for locdata: {locdata}")

    return self



"""

Analyze variance of localization coordinates.

Localization coordinates in localization clusters come with a certain variance.
In resolution-limited clusters the variance is determined by the
localization precision.

A close look at the variance of localization coordinates as function of
localization counts helps to characterize localization clusters [1]_.

References
----------
.. [1] Ebert V, Eiring P, Helmerich DA, Seifert R, Sauer M, Doose S.
   Convex hull as diagnostic tool in single-molecule localization microscopy.
   Bioinformatics 38(24), 2022, 5421-5429, doi: 10.1093/bioinformatics/btac700.

"""

# todo add fit procedure to estimate variances
from __future__ import annotations

import logging
import sys
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import boost_histogram as bh
import matplotlib.pyplot as plt
import pandas as pd

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis
from locan.data.validation import _check_loc_properties
from locan.process.aggregate import Bins
from locan.utils.statistics import biased_variance

if TYPE_CHECKING:
    import matplotlib as mpl

    from locan.data.locdata import LocData

__all__: list[str] = ["PositionVarianceExpectation"]

logger = logging.getLogger(__name__)


# The algorithms


def _property_variances(
    collection: LocData,
    loc_property: str,
    biased: bool = False,
) -> pd.DataFrame:
    assert collection is not None and isinstance(  # type narrowing # noqa: S101
        collection.references, Sequence
    )
    loc_property = _check_loc_properties(collection, loc_properties=loc_property)[0]
    loc_property_var = loc_property + "_var"

    ddof = 0 if biased else 1
    results_df = pd.DataFrame()
    results_df["localization_count"] = collection.data.loc[:, "localization_count"]
    results_df[loc_property_var] = [
        reference_.data[loc_property].var(ddof=ddof)
        for reference_ in collection.references
    ]

    return results_df


# The specific analysis classes


@dataclass(repr=False)
class PositionVarianceExpectationResults:
    values: pd.DataFrame = field(default_factory=pd.DataFrame)
    # with index being reference_index
    # with columns: loc_property, other_loc_property, value_to_expectation_ratio
    grouped: pd.DataFrame = field(default_factory=pd.DataFrame)
    # with index being other_loc_property
    # with columns: loc_property_mean, loc_property_std, expectation


class PositionVarianceExpectation(_Analysis):
    """
    Analyze variation of localization properties
    in relation to expected variations.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    loc_property: str
        The localization property to analyze.
    expectation : int | float | Mapping[str, Any] | pd.Series[Any] | None
        The expected variance for all or each localization property.
        The expected variance equals the squared localization precision
        for localization position coordinates.
    biased : bool
        Flag to use biased or
        unbiased (Bessel-corrected) variance

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : PositionVarianceExpectationResults
        Computed results.
    distribution_statistics : Distribution_stats, None
        Distribution parameters derived from MLE fitting of results.
    """

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        loc_property: str = "position_x",
        expectation: int | float | Mapping[str, Any] | pd.Series[Any] | None = None,
        biased: bool = True,
    ) -> None:
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.results = None
        self.distribution_statistics = None

    def compute(self, locdata: LocData) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data.

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        loc_property = self.parameter["loc_property"] + "_var"
        other_loc_property = "localization_count"

        self.results = PositionVarianceExpectationResults()

        self.results.values = _property_variances(
            collection=locdata,
            loc_property=self.parameter["loc_property"],
            biased=self.parameter["biased"],
        )

        grouped = self.results.values.groupby(other_loc_property)

        self.results.grouped[loc_property + "_mean"] = grouped.mean()
        self.results.grouped[loc_property + "_std"] = grouped.std()

        expectation = self.parameter["expectation"]
        if expectation is None:
            self.results.grouped["expectation"] = pd.NA
        elif isinstance(expectation, (int, float)) and self.parameter["biased"] is True:
            n_samples = self.results.grouped.index
            expectation = biased_variance(variance=expectation, n_samples=n_samples)
            expectation = pd.Series(data=expectation, index=n_samples)
            self.results.grouped["expectation"] = expectation
        elif isinstance(expectation, (int, float)):
            self.results.grouped["expectation"] = expectation
        elif isinstance(expectation, pd.Series):
            indices = self.results.grouped.index
            self.results.grouped["expectation"] = expectation.loc[indices]
        elif isinstance(expectation, Mapping):
            self.results.grouped["expectation"] = [
                expectation[index_] for index_ in self.results.grouped.index
            ]

        self.results.values["expectation"] = self.results.grouped.loc[
            self.results.values[other_loc_property], "expectation"
        ].to_numpy()

        self.results.values["value_to_expectation_ratio"] = (
            self.results.values[loc_property] / self.results.values["expectation"]
        )

        return self

    def plot(self, ax: mpl.axes.Axes | None = None, **kwargs: Any) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        variances as function of localization counts.

        Parameters
        ----------
        ax
            The axes on which to show the image
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        self.results.values.plot(
            kind="scatter",
            alpha=0.2,
            x="localization_count",
            y=self.parameter["loc_property"] + "_var",
            color="gray",
            ax=ax,
            **dict(dict(legend=False), **kwargs),
        )

        self.results.grouped.plot(
            kind="line",
            marker=".",
            x=None,
            y=self.parameter["loc_property"] + "_var_mean",
            color="blue",
            ax=ax,
        )

        if not self.results.grouped.expectation.isna().all():
            self.results.grouped.plot(
                kind="line",
                x=None,
                y="expectation",
                color="black",
                ax=ax,
            )

        ax.set(
            title="Position Variance Expectation",
            xlabel="localization_count",
            ylabel=self.parameter["loc_property"] + "_var",
        )

        return ax

    def hist(
        self,
        ax: mpl.axes.Axes | None = None,
        bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
        n_bins: int | Sequence[int] | None = None,
        bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = None,
        bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
        bin_range: (
            tuple[float, float] | Sequence[float] | Sequence[Sequence[float]] | None
        ) = None,
        log: bool = True,
        fit: bool = False,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        2-dimensional histogram of variances and localization counts.

        Parameters
        ----------
        ax
            The axes on which to show the image
        bins
            The bin specification as defined in :class:`Bins`
        bin_edges
            Bin edges for all or each dimension
            with shape (dimension, n_bin_edges).
        bin_range
            Minimum and maximum edge for all or each dimensions
            with shape (2,) or (dimension, 2).
        n_bins
            The number of bins for all or each dimension.
            5 yields 5 bins in all dimensions.
            (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
        bin_size
            The size of bins for all or each bin and for all or each dimension
            with shape (dimension,) or (dimension, n_bins).
            5 would describe bin_size of 5 for all bins in all dimensions.
            ((2, 5),) yield bins of size (2, 5) for one dimension.
            (2, 5) yields bins of size 2 for one dimension and 5 for the other
            dimension.
            ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
            (1, 3) for the other dimension.
            To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
        log
            Flag for plotting on a log scale.
        fit
            Flag indicating if distribution fit is shown.
            The fit will only be computed if `distribution_statistics`
            is None.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.pcolormesh`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        fig = ax.get_figure()

        other_loc_property = "localization_count"
        loc_property = self.parameter["loc_property"] + "_var"

        if all(
            item_ is None for item_ in [bins, n_bins, bin_size, bin_edges, bin_range]
        ):
            max_other_loc_property_ = self.results.grouped.index.max()
            max_loc_property_ = self.results.values[loc_property].max()

            if log:
                axes = bh.axis.AxesTuple(
                    (
                        bh.axis.Regular(
                            max_other_loc_property_,
                            1,
                            max_other_loc_property_,
                            transform=bh.axis.transform.log,
                        ),
                        bh.axis.Regular(
                            50, 1, max_loc_property_, transform=bh.axis.transform.log
                        ),
                    )
                )
            else:
                axes = bh.axis.AxesTuple(
                    (
                        bh.axis.Regular(
                            max_other_loc_property_, 1, max_other_loc_property_
                        ),
                        bh.axis.Regular(50, 1, max_loc_property_),
                    )
                )
            bins = Bins(bins=axes)
        else:
            try:
                bins = Bins(
                    bins,
                    n_bins,
                    bin_size,
                    bin_edges,
                    bin_range,
                    labels=[loc_property, other_loc_property],
                )
            except ValueError as exc:
                # todo: check if message is appropriate
                # the error is raised again only to adapt the message.
                raise ValueError(
                    "Bin dimension and len of `loc_properties` is incompatible."
                ) from exc
        axes = bins.boost_histogram_axes
        histogram = bh.Histogram(*axes)
        histogram.reset()
        histogram.fill(
            self.results.values[other_loc_property], self.results.values[loc_property]
        )
        mesh = ax.pcolormesh(*histogram.axes.edges.T, histogram.view().T, **kwargs)
        fig.colorbar(mesh)  # type:ignore[union-attr]

        self.results.grouped.plot(
            kind="line",
            marker=".",
            x=None,
            y=self.parameter["loc_property"] + "_var_mean",
            color="blue",
            ax=ax,
        )

        if not self.results.grouped.expectation.isna().all():
            self.results.grouped.plot(
                kind="line",
                x=None,
                y="expectation",
                color="white",
                ax=ax,
            )

        ax.set(
            title="Position Variance Expectation",
            xlabel=other_loc_property,
            ylabel=loc_property,
        )

        if log:
            ax.set(xscale="log", yscale="log")

        if fit:
            raise NotImplementedError

        return ax



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class Source(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    UNKNOWN_SOURCE: _ClassVar[Source]
    DESIGN: _ClassVar[Source]
    EXPERIMENT: _ClassVar[Source]
    SIMULATION: _ClassVar[Source]
    IMPORT: _ClassVar[Source]

class State(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    UNKNOWN_STATE: _ClassVar[State]
    RAW: _ClassVar[State]
    MODIFIED: _ClassVar[State]

class File_type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    UNKNOWN_FILE_TYPE: _ClassVar[File_type]
    CUSTOM: _ClassVar[File_type]
    RAPIDSTORM: _ClassVar[File_type]
    ELYRA: _ClassVar[File_type]
    THUNDERSTORM: _ClassVar[File_type]
    ASDF: _ClassVar[File_type]
    NANOIMAGER: _ClassVar[File_type]
    RAPIDSTORMTRACK: _ClassVar[File_type]
    SMLM: _ClassVar[File_type]
    DECODE: _ClassVar[File_type]
    SMAP: _ClassVar[File_type]

UNKNOWN_SOURCE: Source
DESIGN: Source
EXPERIMENT: Source
SIMULATION: Source
IMPORT: Source
UNKNOWN_STATE: State
RAW: State
MODIFIED: State
UNKNOWN_FILE_TYPE: File_type
CUSTOM: File_type
RAPIDSTORM: File_type
ELYRA: File_type
THUNDERSTORM: File_type
ASDF: File_type
NANOIMAGER: File_type
RAPIDSTORMTRACK: File_type
SMLM: File_type
DECODE: File_type
SMAP: File_type

class Operation(_message.Message):
    __slots__ = ("name", "parameter")
    NAME_FIELD_NUMBER: _ClassVar[int]
    PARAMETER_FIELD_NUMBER: _ClassVar[int]
    name: str
    parameter: str
    def __init__(
        self, name: _Optional[str] = ..., parameter: _Optional[str] = ...
    ) -> None: ...

class File(_message.Message):
    __slots__ = ("identifier", "comment", "type", "path", "groups")
    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    PATH_FIELD_NUMBER: _ClassVar[int]
    GROUPS_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    type: File_type
    path: str
    groups: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        type: _Optional[_Union[File_type, str]] = ...,
        path: _Optional[str] = ...,
        groups: _Optional[_Iterable[str]] = ...,
    ) -> None: ...

class Address(_message.Message):
    __slots__ = ("address_lines", "city", "city_code", "country")
    ADDRESS_LINES_FIELD_NUMBER: _ClassVar[int]
    CITY_FIELD_NUMBER: _ClassVar[int]
    CITY_CODE_FIELD_NUMBER: _ClassVar[int]
    COUNTRY_FIELD_NUMBER: _ClassVar[int]
    address_lines: _containers.RepeatedScalarFieldContainer[str]
    city: str
    city_code: str
    country: str
    def __init__(
        self,
        address_lines: _Optional[_Iterable[str]] = ...,
        city: _Optional[str] = ...,
        city_code: _Optional[str] = ...,
        country: _Optional[str] = ...,
    ) -> None: ...

class Affiliation(_message.Message):
    __slots__ = ("institute", "department", "address")
    INSTITUTE_FIELD_NUMBER: _ClassVar[int]
    DEPARTMENT_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    institute: str
    department: str
    address: Address
    def __init__(
        self,
        institute: _Optional[str] = ...,
        department: _Optional[str] = ...,
        address: _Optional[_Union[Address, _Mapping]] = ...,
    ) -> None: ...

class Person(_message.Message):
    __slots__ = (
        "identifier",
        "comment",
        "first_name",
        "last_name",
        "title",
        "affiliations",
        "address",
        "emails",
        "roles",
    )
    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    FIRST_NAME_FIELD_NUMBER: _ClassVar[int]
    LAST_NAME_FIELD_NUMBER: _ClassVar[int]
    TITLE_FIELD_NUMBER: _ClassVar[int]
    AFFILIATIONS_FIELD_NUMBER: _ClassVar[int]
    ADDRESS_FIELD_NUMBER: _ClassVar[int]
    EMAILS_FIELD_NUMBER: _ClassVar[int]
    ROLES_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    first_name: str
    last_name: str
    title: str
    affiliations: _containers.RepeatedCompositeFieldContainer[Affiliation]
    address: Address
    emails: _containers.RepeatedScalarFieldContainer[str]
    roles: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        first_name: _Optional[str] = ...,
        last_name: _Optional[str] = ...,
        title: _Optional[str] = ...,
        affiliations: _Optional[_Iterable[_Union[Affiliation, _Mapping]]] = ...,
        address: _Optional[_Union[Address, _Mapping]] = ...,
        emails: _Optional[_Iterable[str]] = ...,
        roles: _Optional[_Iterable[str]] = ...,
    ) -> None: ...

class ExperimentalSample(_message.Message):
    __slots__ = ("identifier", "comment", "targets", "fluorophores", "buffers", "map")

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    TARGETS_FIELD_NUMBER: _ClassVar[int]
    FLUOROPHORES_FIELD_NUMBER: _ClassVar[int]
    BUFFERS_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    targets: _containers.RepeatedScalarFieldContainer[str]
    fluorophores: _containers.RepeatedScalarFieldContainer[str]
    buffers: _containers.RepeatedScalarFieldContainer[str]
    map: _containers.ScalarMap[str, str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        targets: _Optional[_Iterable[str]] = ...,
        fluorophores: _Optional[_Iterable[str]] = ...,
        buffers: _Optional[_Iterable[str]] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class ExperimentalSetup(_message.Message):
    __slots__ = ("identifier", "comment", "optical_units", "map")

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    OPTICAL_UNITS_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    optical_units: _containers.RepeatedCompositeFieldContainer[OpticalUnit]
    map: _containers.ScalarMap[str, str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        optical_units: _Optional[_Iterable[_Union[OpticalUnit, _Mapping]]] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class OpticalUnit(_message.Message):
    __slots__ = (
        "identifier",
        "comment",
        "illumination",
        "detection",
        "acquisition",
        "lightsheet",
    )
    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    ILLUMINATION_FIELD_NUMBER: _ClassVar[int]
    DETECTION_FIELD_NUMBER: _ClassVar[int]
    ACQUISITION_FIELD_NUMBER: _ClassVar[int]
    LIGHTSHEET_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    illumination: Illumination
    detection: Detection
    acquisition: Acquisition
    lightsheet: Lightsheet
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        illumination: _Optional[_Union[Illumination, _Mapping]] = ...,
        detection: _Optional[_Union[Detection, _Mapping]] = ...,
        acquisition: _Optional[_Union[Acquisition, _Mapping]] = ...,
        lightsheet: _Optional[_Union[Lightsheet, _Mapping]] = ...,
    ) -> None: ...

class Illumination(_message.Message):
    __slots__ = (
        "identifier",
        "comment",
        "lightsource",
        "power",
        "area",
        "power_density",
        "wavelength",
        "map",
    )

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    LIGHTSOURCE_FIELD_NUMBER: _ClassVar[int]
    POWER_FIELD_NUMBER: _ClassVar[int]
    AREA_FIELD_NUMBER: _ClassVar[int]
    POWER_DENSITY_FIELD_NUMBER: _ClassVar[int]
    WAVELENGTH_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    lightsource: str
    power: float
    area: float
    power_density: float
    wavelength: float
    map: _containers.ScalarMap[str, str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        lightsource: _Optional[str] = ...,
        power: _Optional[float] = ...,
        area: _Optional[float] = ...,
        power_density: _Optional[float] = ...,
        wavelength: _Optional[float] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class Detection(_message.Message):
    __slots__ = ("identifier", "comment", "camera", "map")

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    CAMERA_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    camera: Camera
    map: _containers.ScalarMap[str, str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        camera: _Optional[_Union[Camera, _Mapping]] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class Camera(_message.Message):
    __slots__ = (
        "identifier",
        "comment",
        "name",
        "model",
        "gain",
        "electrons_per_count",
        "integration_time",
        "pixel_count_x",
        "pixel_count_y",
        "pixel_size_x",
        "pixel_size_y",
        "flipped",
        "map",
        "offset",
        "serial_number",
    )

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    MODEL_FIELD_NUMBER: _ClassVar[int]
    GAIN_FIELD_NUMBER: _ClassVar[int]
    ELECTRONS_PER_COUNT_FIELD_NUMBER: _ClassVar[int]
    INTEGRATION_TIME_FIELD_NUMBER: _ClassVar[int]
    PIXEL_COUNT_X_FIELD_NUMBER: _ClassVar[int]
    PIXEL_COUNT_Y_FIELD_NUMBER: _ClassVar[int]
    PIXEL_SIZE_X_FIELD_NUMBER: _ClassVar[int]
    PIXEL_SIZE_Y_FIELD_NUMBER: _ClassVar[int]
    FLIPPED_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    SERIAL_NUMBER_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    name: str
    model: str
    gain: float
    electrons_per_count: float
    integration_time: _duration_pb2.Duration
    pixel_count_x: int
    pixel_count_y: int
    pixel_size_x: float
    pixel_size_y: float
    flipped: bool
    map: _containers.ScalarMap[str, str]
    offset: float
    serial_number: str
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        name: _Optional[str] = ...,
        model: _Optional[str] = ...,
        gain: _Optional[float] = ...,
        electrons_per_count: _Optional[float] = ...,
        integration_time: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        pixel_count_x: _Optional[int] = ...,
        pixel_count_y: _Optional[int] = ...,
        pixel_size_x: _Optional[float] = ...,
        pixel_size_y: _Optional[float] = ...,
        flipped: bool = ...,
        map: _Optional[_Mapping[str, str]] = ...,
        offset: _Optional[float] = ...,
        serial_number: _Optional[str] = ...,
    ) -> None: ...

class Acquisition(_message.Message):
    __slots__ = (
        "identifier",
        "comment",
        "frame_count",
        "frame_of_interest_first",
        "frame_of_interest_last",
        "time_start",
        "time_end",
        "stack_count",
        "stack_step_count",
        "stack_step_size",
        "map",
    )

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    FRAME_COUNT_FIELD_NUMBER: _ClassVar[int]
    FRAME_OF_INTEREST_FIRST_FIELD_NUMBER: _ClassVar[int]
    FRAME_OF_INTEREST_LAST_FIELD_NUMBER: _ClassVar[int]
    TIME_START_FIELD_NUMBER: _ClassVar[int]
    TIME_END_FIELD_NUMBER: _ClassVar[int]
    STACK_COUNT_FIELD_NUMBER: _ClassVar[int]
    STACK_STEP_COUNT_FIELD_NUMBER: _ClassVar[int]
    STACK_STEP_SIZE_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    frame_count: int
    frame_of_interest_first: int
    frame_of_interest_last: int
    time_start: _timestamp_pb2.Timestamp
    time_end: _timestamp_pb2.Timestamp
    stack_count: int
    stack_step_count: int
    stack_step_size: float
    map: _containers.ScalarMap[str, str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        frame_count: _Optional[int] = ...,
        frame_of_interest_first: _Optional[int] = ...,
        frame_of_interest_last: _Optional[int] = ...,
        time_start: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        time_end: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        stack_count: _Optional[int] = ...,
        stack_step_count: _Optional[int] = ...,
        stack_step_size: _Optional[float] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class Lightsheet(_message.Message):
    __slots__ = ("identifier", "comment", "angle_x", "angle_y", "angle_z", "map")

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    ANGLE_X_FIELD_NUMBER: _ClassVar[int]
    ANGLE_Y_FIELD_NUMBER: _ClassVar[int]
    ANGLE_Z_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    angle_x: float
    angle_y: float
    angle_z: float
    map: _containers.ScalarMap[str, str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        angle_x: _Optional[float] = ...,
        angle_y: _Optional[float] = ...,
        angle_z: _Optional[float] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class Experiment(_message.Message):
    __slots__ = ("identifier", "comment", "experimenters", "samples", "setups", "map")

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    EXPERIMENTERS_FIELD_NUMBER: _ClassVar[int]
    SAMPLES_FIELD_NUMBER: _ClassVar[int]
    SETUPS_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    experimenters: _containers.RepeatedCompositeFieldContainer[Person]
    samples: _containers.RepeatedCompositeFieldContainer[ExperimentalSample]
    setups: _containers.RepeatedCompositeFieldContainer[ExperimentalSetup]
    map: _containers.ScalarMap[str, str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        experimenters: _Optional[_Iterable[_Union[Person, _Mapping]]] = ...,
        samples: _Optional[_Iterable[_Union[ExperimentalSample, _Mapping]]] = ...,
        setups: _Optional[_Iterable[_Union[ExperimentalSetup, _Mapping]]] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class Localizer(_message.Message):
    __slots__ = (
        "identifier",
        "comment",
        "software",
        "intensity_threshold",
        "psf_fixed",
        "psf_size",
        "map",
    )

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    SOFTWARE_FIELD_NUMBER: _ClassVar[int]
    INTENSITY_THRESHOLD_FIELD_NUMBER: _ClassVar[int]
    PSF_FIXED_FIELD_NUMBER: _ClassVar[int]
    PSF_SIZE_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    software: str
    intensity_threshold: float
    psf_fixed: bool
    psf_size: float
    map: _containers.ScalarMap[str, str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        software: _Optional[str] = ...,
        intensity_threshold: _Optional[float] = ...,
        psf_fixed: bool = ...,
        psf_size: _Optional[float] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class Relation(_message.Message):
    __slots__ = ("identifier", "comment", "file", "map")

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    FILE_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    file: File
    map: _containers.ScalarMap[str, str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        file: _Optional[_Union[File, _Mapping]] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class Property(_message.Message):
    __slots__ = ("identifier", "comment", "name", "unit", "type", "map")

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    UNIT_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    name: str
    unit: str
    type: str
    map: _containers.ScalarMap[str, str]
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        name: _Optional[str] = ...,
        unit: _Optional[str] = ...,
        type: _Optional[str] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
    ) -> None: ...

class Metadata(_message.Message):
    __slots__ = (
        "identifier",
        "comment",
        "source",
        "state",
        "history",
        "ancestor_identifiers",
        "properties",
        "localization_properties",
        "element_count",
        "frame_count",
        "file",
        "relations",
        "experiment",
        "localizer",
        "map",
        "creation_time",
        "modification_time",
        "production_time",
    )

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    IDENTIFIER_FIELD_NUMBER: _ClassVar[int]
    COMMENT_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    HISTORY_FIELD_NUMBER: _ClassVar[int]
    ANCESTOR_IDENTIFIERS_FIELD_NUMBER: _ClassVar[int]
    PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    LOCALIZATION_PROPERTIES_FIELD_NUMBER: _ClassVar[int]
    ELEMENT_COUNT_FIELD_NUMBER: _ClassVar[int]
    FRAME_COUNT_FIELD_NUMBER: _ClassVar[int]
    FILE_FIELD_NUMBER: _ClassVar[int]
    RELATIONS_FIELD_NUMBER: _ClassVar[int]
    EXPERIMENT_FIELD_NUMBER: _ClassVar[int]
    LOCALIZER_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    CREATION_TIME_FIELD_NUMBER: _ClassVar[int]
    MODIFICATION_TIME_FIELD_NUMBER: _ClassVar[int]
    PRODUCTION_TIME_FIELD_NUMBER: _ClassVar[int]
    identifier: str
    comment: str
    source: Source
    state: State
    history: _containers.RepeatedCompositeFieldContainer[Operation]
    ancestor_identifiers: _containers.RepeatedScalarFieldContainer[str]
    properties: _containers.RepeatedCompositeFieldContainer[Property]
    localization_properties: _containers.RepeatedCompositeFieldContainer[Property]
    element_count: int
    frame_count: int
    file: File
    relations: _containers.RepeatedCompositeFieldContainer[Relation]
    experiment: Experiment
    localizer: Localizer
    map: _containers.ScalarMap[str, str]
    creation_time: _timestamp_pb2.Timestamp
    modification_time: _timestamp_pb2.Timestamp
    production_time: _timestamp_pb2.Timestamp
    def __init__(
        self,
        identifier: _Optional[str] = ...,
        comment: _Optional[str] = ...,
        source: _Optional[_Union[Source, str]] = ...,
        state: _Optional[_Union[State, str]] = ...,
        history: _Optional[_Iterable[_Union[Operation, _Mapping]]] = ...,
        ancestor_identifiers: _Optional[_Iterable[str]] = ...,
        properties: _Optional[_Iterable[_Union[Property, _Mapping]]] = ...,
        localization_properties: _Optional[_Iterable[_Union[Property, _Mapping]]] = ...,
        element_count: _Optional[int] = ...,
        frame_count: _Optional[int] = ...,
        file: _Optional[_Union[File, _Mapping]] = ...,
        relations: _Optional[_Iterable[_Union[Relation, _Mapping]]] = ...,
        experiment: _Optional[_Union[Experiment, _Mapping]] = ...,
        localizer: _Optional[_Union[Localizer, _Mapping]] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
        creation_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        modification_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        production_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...



"""

Visualize localization data.

This module provides functions to visualize localization data
with matplotlib or napari.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   colormap
   render
   render_mpl
   render_napari
   transform

"""

from __future__ import annotations

from locan.visualize import colormap, render, render_mpl, render_napari, transform

from .colormap import *
from .render import *
from .render_mpl import (
    apply_window as apply_window,
    render_2d_mpl as render_2d_mpl,
    render_2d_rgb_mpl as render_2d_rgb_mpl,
    render_2d_scatter_density as render_2d_scatter_density,
    scatter_2d_mpl as scatter_2d_mpl,
    scatter_3d_mpl as scatter_3d_mpl,
)
from .render_napari import (
    get_rois as get_rois,
    render_2d_napari as render_2d_napari,
    render_2d_napari_image as render_2d_napari_image,
    render_2d_rgb_napari as render_2d_rgb_napari,
    render_3d_napari as render_3d_napari,
    render_3d_napari_image as render_3d_napari_image,
    render_3d_rgb_napari as render_3d_rgb_napari,
    save_rois as save_rois,
    select_by_drawing_napari as select_by_drawing_napari,
)
from .transform import *

__all__: list[str] = []
__all__.extend(colormap.__all__)
__all__.extend(render.__all__)
__all__.extend(render_mpl.__all__)
__all__.extend(render_napari.__all__)
__all__.extend(transform.__all__)



from io import StringIO

import locan.constants
from locan import FileType
from locan.locan_io import load_thunderstorm_file
from locan.locan_io.locdata.thunderstorm_io import load_thunderstorm_header


def test_get_correct_column_names_from_Thunderstorm_header():
    columns = load_thunderstorm_header(
        path=locan.ROOT_DIR / "tests/test_data/Thunderstorm_dstorm_data.csv"
    )
    assert columns == [
        "original_index",
        "frame",
        "position_x",
        "position_y",
        "psf_sigma",
        "intensity",
        "local_background",
        "local_background_sigma",
        "chi_square",
        "uncertainty",
    ]

    file_like = StringIO(
        "id,frame,x [nm],y [nm]\n73897.0,2001.0,1320.109670647555,26344.7124618434"
    )
    columns = load_thunderstorm_header(path=file_like)
    assert columns == ["original_index", "frame", "position_x", "position_y"]


def test_loading_Thunderstorm_file():
    file_path = locan.ROOT_DIR / "tests/test_data/Thunderstorm_dstorm_data.csv"
    dat = load_thunderstorm_file(path=file_path, nrows=10)
    # print(dat.data.columns)
    assert len(dat) == 10
    assert dat.meta.file.type == FileType.THUNDERSTORM.value
    assert dat.meta.file.path == str(file_path)

    file_like = StringIO(
        "id,frame,x [nm],y [nm]\n73897.0,2001.0,1320.109670647555,26344.7124618434"
    )
    dat = load_thunderstorm_file(path=file_like, nrows=1)
    assert len(dat) == 1



"""
Benchmark functions for :mod:`locan.simulation`
"""

import numpy as np

import locan as lc

rng = np.random.default_rng(seed=1)


class BenchmarkResample:
    """
    Benchmarks for data.metadata_pb2.Metadata objects
    """

    def __init__(self):
        self.locdata = None
        self.new_locdata = None

    def setup(self):
        self.locdata = lc.simulate_uniform(n_samples=1000, seed=rng)
        self.locdata.data["uncertainty"] = 0.01
        self.locdata.data["other_property"] = 111

    def time_resample(self):
        self.new_locdata = lc.resample(self.locdata, n_samples=1000, seed=rng)

    def mem_locdata(self):
        return self.new_locdata


def main():
    bm = BenchmarkResample()
    bm.setup()
    print(bm.locdata.data.info())
    bm.time_resample()
    print(bm.new_locdata.data.info())


def main_profile():
    """
    Run benchmarks through profiler.
    """
    import cProfile
    import pstats

    bm = BenchmarkResample()
    bm.setup()

    profiler = cProfile.Profile()
    profiler.enable()

    for _ in range(10):
        bm.time_resample()

    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats("cumtime")
    stats.print_stats("time_")


if __name__ == "__main__":
    main()
    main_profile()



"""

Validate localization data.

This module provides functions to validate LocData properties.

"""

from __future__ import annotations

from collections.abc import Iterable
from typing import Protocol

import pandas as pd

from locan.data.locdata import LocData

__all__: list[str] = []


class Collection(Protocol):
    data: pd.DataFrame
    references: Iterable[LocData]
    coordinate_keys: list[str]

    def __len__(self) -> int: ...


def _check_loc_properties(
    locdata: LocData | Collection, loc_properties: str | Iterable[str] | None = None
) -> list[str]:
    """
    Check that loc_properties are valid properties in locdata.

    Parameters
    ----------
    locdata
        Localization data
    loc_properties
        LocData property names. If None coordinate labels are used.

    Returns
    -------
    list[str]
        Valid localization property names
    """
    if loc_properties is None:  # use coordinate_labels
        labels = locdata.coordinate_keys.copy()
    elif isinstance(loc_properties, str):
        if loc_properties not in locdata.data.columns:
            raise ValueError(
                f"{loc_properties} is not a valid property in locdata.data."
            )
        labels = [loc_properties]
    elif isinstance(loc_properties, (tuple, list)):
        labels = list(loc_properties)
        for loc_property in loc_properties:
            if loc_property not in locdata.data.columns:
                raise ValueError(
                    f"{loc_property} is not a valid property in locdata.data."
                )
    else:
        raise ValueError(f"{loc_properties} is not a valid property in locdata.data.")
    return labels



import os
from pathlib import Path

import pytest

from locan import ROOT_DIR
from locan.dependencies import HAS_DEPENDENCY
from locan.scripts.script_napari import sc_napari


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_script_napari():
    path = Path(ROOT_DIR / "tests/test_data/five_blobs.txt")
    assert path.exists()
    sc_napari(file_path=str(path), file_type=1, bin_size=20, rescale=None)


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_script_napari_from_sys():
    path = Path(ROOT_DIR / "tests/test_data/five_blobs.txt")
    assert path.exists()
    exit_status = os.system(
        f"locan napari -f {str(path)} -t 1 --bin_size 20 --rescale None"
    )
    assert exit_status == 0


# pass to CL:
# locan napari -f "locan/tests/test_data/five_blobs.txt" -t 1 --bin_size 20 --rescale None



"""
Drift analysis for localization coordinates.

This module provides functions for estimating spatial drift in localization
data.

Software based drift correction using image correlation has been described in
several publications .
Methods employed for drift estimation comprise single molecule localization
analysis (an iterative closest point (icp)
algorithm as implemented in the open3d library [1]_, [2]_) or image
cross-correlation analysis [3]_, [4]_, [5]_.

Examples
--------
Please use the following procedure to estimate and correct for spatial drift::

    from lmfit import LinearModel
    drift = Drift(chunk_size=1000, target='first').\\
            compute(locdata).\\
            fit_transformations(slice_data=slice(0, -1),
                                matrix_models=None,
                                offset_models=(LinearModel(), LinearModel())).\\
            apply_correction()
    locdata_corrected = drift.locdata_corrected

References
----------
.. [1] Qian-Yi Zhou, Jaesik Park, Vladlen Koltun,
   Open3D: A Modern Library for 3D Data Processing,
   arXiv 2018, 1801.09847

.. [2] Rusinkiewicz and M. Levoy,
   Efficient variants of the ICP algorithm,
   In 3-D Digital Imaging and Modeling, 2001.

.. [3] C. Geisler,
   Drift estimation for single marker switching based imaging schemes,
   Optics Express. 2012, 20(7):7274-89.

.. [4] Yina Wang et al.,
   Localization events-based sample drift correction for localization
   microscopy with redundant cross-correlation
   algorithm, Optics Express 2014, 22(13):15982-91.

.. [5] Michael J. Mlodzianoski et al.,
   Sample drift correction in 3D fluorescence photoactivation localization
   microscopy,
   Opt Express. 2011 Aug 1;19(16):15009-19.

"""

from __future__ import annotations

import logging
import sys
from collections.abc import Sequence
from typing import Any, Literal, Protocol

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
from lmfit.model import ModelResult as ModelResultLmf
from lmfit.models import ConstantModel, LinearModel, PolynomialModel
from lmfit.models import Model as ModelLmf
from scipy.interpolate import splev, splrep

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis
from locan.data.locdata import LocData
from locan.data.metadata_utils import _modify_meta
from locan.dependencies import needs_package
from locan.process.register import Transformation, _register_icp_open3d, register_cc
from locan.process.transform.spatial_transformation import transform_affine

__all__: list[str] = ["Drift", "DriftComponent"]

logger = logging.getLogger(__name__)


class DriftModel(Protocol):
    def fit(self, *args: Any, **kwargs: Any) -> Any: ...

    def eval(self, *args: Any, **kwargs: Any) -> Any: ...

    def plot(self, *args: Any, **kwargs: Any) -> Any: ...


# The algorithms


@needs_package("open3d")
def _estimate_drift_icp(
    locdata: LocData,
    chunks: Sequence[tuple[int, ...]] | None = None,
    chunk_size: int | None = None,
    n_chunks: int | None = None,
    target: Literal["first", "previous"] = "first",
    kwargs_chunk: dict[str, Any] | None = None,
    kwargs_register: dict[str, Any] | None = None,
) -> tuple[LocData, list[Transformation]]:
    """
    Estimate drift from localization coordinates by registering points in
    successive time-chunks of localization
    data using an "Iterative Closest Point" algorithm.

    Parameters
    ----------
    locdata
       Localization data with properties for coordinates and frame.
    chunks
        Localization chunks as defined by a list of index-tuples.
        One of `chunks`, `chunk_size` or `n_chunks` must be different
        from None.
    chunk_size
        Number of consecutive localizations to form a single chunk of data.
        One of `chunks`, `chunk_size` or `n_chunks` must be different
        from None.
    n_chunks
        Number of chunks.
        One of `chunks`, `chunk_size` or `n_chunks` must be different
        from None.
    target
       The chunk on which all other chunks are aligned. One of 'first', 'previous'.
    kwargs_chunk
        Other parameter passed to :meth:`LocData.from_chunks`.
    kwargs_register
        Other parameter passed to :func:`_register_icp_open3d`.

    Returns
    -------
    tuple[LocData, list[Transformation]]
        Collection and corresponding transformations.
    """
    if kwargs_chunk is None:
        kwargs_chunk = {}
    if kwargs_register is None:
        kwargs_register = {}

    # split in chunks
    collection = LocData.from_chunks(
        locdata, chunks=chunks, chunk_size=chunk_size, n_chunks=n_chunks, **kwargs_chunk
    )
    assert isinstance(collection.references, Sequence)  # type narrowing # noqa: S101

    # register locdatas
    # initialize with identity transformation for chunk zero.
    transformations = [
        Transformation(np.identity(locdata.dimension), np.zeros(locdata.dimension))
    ]
    if target == "first":
        for locdata in collection.references[1:]:
            transformation = _register_icp_open3d(
                locdata.coordinates,
                collection.references[0].coordinates,
                **dict(
                    dict(
                        matrix=None,
                        offset=None,
                        pre_translation=None,
                        max_correspondence_distance=100,
                        max_iteration=10_000,
                        verbose=False,
                    ),
                    **kwargs_register,
                ),
            )
            transformations.append(transformation)

    elif target == "previous":
        for n in range(len(collection.references) - 1):
            transformation = _register_icp_open3d(
                collection.references[n + 1].coordinates,
                collection.references[n].coordinates,
                **dict(
                    dict(
                        matrix=None,
                        offset=None,
                        pre_translation=None,
                        max_correspondence_distance=100,
                        max_iteration=10_000,
                        with_scaling=False,
                        verbose=False,
                    ),
                    **kwargs_register,
                ),
            )
            transformations.append(transformation)

    return collection, transformations


def _estimate_drift_cc(
    locdata: LocData,
    chunks: Sequence[tuple[int, ...]] | None = None,
    chunk_size: int | None = None,
    n_chunks: int | None = None,
    target: Literal["first", "previous"] = "first",
    bin_size: int | float | tuple[int | float] = 10,
    kwargs_chunk: dict[str, Any] | None = None,
    kwargs_register: dict[str, Any] | None = None,
) -> tuple[LocData, list[Transformation]]:
    """
    Estimate drift from localization coordinates by registering points in
    successive time-chunks of localization
    data using a cross-correlation algorithm.

    Parameters
    ----------
    locdata
       Localization data with properties for coordinates and frame.
    chunks
        Localization chunks as defined by a list of index-tuples.
        One of `chunks`, `chunk_size` or `n_chunks` must be different
        from None.
    chunk_size
        Number of consecutive localizations to form a single chunk of data.
        One of `chunks`, `chunk_size` or `n_chunks` must be different
        from None.
    n_chunks
        Number of chunks.
        One of `chunks`, `chunk_size` or `n_chunks` must be different
        from None.
    target
       The chunk on which all other chunks are aligned. One of 'first', 'previous'.
    bin_size
        Size per image pixel
    kwargs_chunk
        Other parameter passed to :meth:`LocData.from_chunks`.
    kwargs_register
        Other parameter passed to :func:`register_cc`.

    Returns
    -------
    tuple[LocData, list[Transformation]]
        Collection and corresponding transformations.
    """
    if kwargs_chunk is None:
        kwargs_chunk = {}
    if kwargs_register is None:
        kwargs_register = {}

    # split in chunks
    collection = LocData.from_chunks(
        locdata, chunks=chunks, chunk_size=chunk_size, n_chunks=n_chunks, **kwargs_chunk
    )
    assert isinstance(collection.references, Sequence)  # type narrowing # noqa: S101
    assert locdata.bounding_box is not None  # type narrowing # noqa: S101
    ranges = locdata.bounding_box.hull.T

    # register images
    # initialize with identity transformation for chunk zero.
    transformations = [
        Transformation(np.identity(locdata.dimension), np.zeros(locdata.dimension))
    ]
    if target == "first":
        for reference in collection.references[1:]:
            transformation = register_cc(
                reference,
                collection.references[0],
                **dict(dict(bin_range=ranges, bin_size=bin_size), **kwargs_register),
            )
            transformations.append(transformation)

    elif target == "previous":
        for n in range(len(collection) - 1):
            transformation = register_cc(
                collection.references[n + 1],
                collection.references[n],
                **dict(dict(bin_range=ranges, bin_size=bin_size), **kwargs_register),
            )
            transformations.append(transformation)

    return collection, transformations


# The specific analysis classes


class _LmfitModelFacade:
    def __init__(self, model: ModelLmf) -> None:
        self.model: ModelLmf = model
        self.model_result: ModelResultLmf | None = None

    def fit(
        self, x: npt.ArrayLike, y: npt.ArrayLike, verbose: bool = False, **kwargs: Any
    ) -> ModelResultLmf:
        params = self.model.guess(data=y, x=x)
        self.model_result = self.model.fit(x=x, data=y, params=params, **kwargs)
        if verbose:
            self.plot()
        return self.model_result

    def eval(self, x: npt.ArrayLike) -> npt.NDArray[np.float64]:
        x = np.asarray(x)
        if self.model_result is None:
            raise AttributeError("No model_result available. Run fit method first.")
        return_value: npt.NDArray[np.float64] = self.model_result.eval(x=x)
        return return_value

    def plot(self, **kwargs: Any) -> mpl.axes.Axes:
        if self.model_result is None:
            raise AttributeError("No model_result available. Run fit method first.")
        return_value: mpl.axes.Axes = self.model_result.plot(**kwargs)
        return return_value


class _ConstantModelFacade:
    def __init__(self, **kwargs: Any) -> None:
        self.model: ModelLmf = ConstantModel(**kwargs)
        self.model_result: ModelResultLmf | None = None
        self.independent_variable: npt.NDArray[Any] | None = None

    def fit(
        self, x: npt.ArrayLike, y: npt.ArrayLike, verbose: bool = False, **kwargs: Any
    ) -> ModelResultLmf:
        self.independent_variable = np.asarray(x)
        self.model_result = self.model.fit(x=x, data=y, **kwargs)
        if verbose:
            self.plot()
        return self.model_result

    def eval(self, x: npt.ArrayLike) -> npt.NDArray[np.float64]:
        x = np.asarray(x)
        if self.model_result is None:
            raise AttributeError("No model_result available. Run fit method first.")
        result: npt.NDArray[np.float64] = self.model_result.eval(x=x)
        if np.shape(result) != np.shape(x):  # needed to work with lmfit<1.2.0
            result = np.full(shape=np.shape(x), fill_value=result)
        return result

    def plot(self, **kwargs: Any) -> mpl.axes.Axes:
        if self.model_result is None:
            raise AttributeError("No model_result available. Run fit method first.")
        assert self.independent_variable is not None  # type narrowing # noqa: S101
        assert self.model_result.data is not None  # type narrowing # noqa: S101
        x = self.independent_variable
        y = self.model_result.data
        return_value: mpl.axes.Axes = plt.plot(x, y, "o", x, self.eval(x=x), **kwargs)  # type: ignore[assignment]
        return return_value


class _ConstantZeroModelFacade(_ConstantModelFacade):
    def __init__(self) -> None:
        self.model: ModelLmf = ConstantModel()
        self.model.set_param_hint(name="c", value=0, vary=False)
        self.model_result: ModelResultLmf | None = None


class _ConstantOneModelFacade(_ConstantModelFacade):
    def __init__(self) -> None:
        self.model: ModelLmf = ConstantModel()
        self.model.set_param_hint(name="c", value=1, vary=False)
        self.model_result: ModelResultLmf | None = None


class _SplineModelFacade:
    def __init__(self, **kwargs: Any) -> None:
        self.model: str = "spline"
        self.model_result: tuple[Any, Any, Any] | None = None
        self.parameter = kwargs
        self.independent_variable: npt.NDArray[Any] | None = None
        self.data: npt.NDArray[np.float64] | None = None

    def fit(
        self, x: npt.ArrayLike, y: npt.ArrayLike, verbose: bool = False, **kwargs: Any
    ) -> tuple[Any, Any, Any]:
        self.independent_variable = np.asarray(x)
        self.data = np.asarray(y)
        self.model_result = splrep(
            x, y, **dict(dict(k=3, s=100), **dict(**self.parameter, **kwargs))
        )
        if verbose:
            self.plot()
        assert self.model_result is not None  # type narrowing # noqa: S101
        return self.model_result

    def eval(self, x: npt.ArrayLike) -> float | npt.NDArray[np.float64] | list[Any]:
        x = np.asarray(x)
        if self.model_result is None:
            raise AttributeError("No model_result available. Run fit method first.")
        results: npt.NDArray[np.float64] | list[Any] = splev(x, self.model_result)
        if isinstance(x, (tuple, list, np.ndarray)):
            return results
        else:
            return float(results)

    def plot(self, **kwargs: Any) -> mpl.axes.Axes:
        if self.model_result is None:
            raise AttributeError("No model_result available. Run fit method first.")
        x = self.independent_variable
        y = self.data
        x_ = np.linspace(np.min(x), np.max(x), 100)  # type: ignore[arg-type]

        return_value: mpl.axes.Axes = plt.plot(x, y, "o", x_, self.eval(x=x_), **kwargs)  # type: ignore
        return return_value


class DriftComponent:
    """
    Class carrying model functions to describe drift over time
    (in unit of frames).
    DriftComponent provides a transformation to apply a drift correction.

    Standard models for constant, linear or polynomial drift correction are
    taken from :mod:`lmfit.models`.
    For fitting splines we use the scipy function :func:`scipy.interpolate.splrep`.

    Parameters
    ----------
    type : Literal["none", "zero", "one", "constant", "linear", "polynomial", "spline"] | lmfit.models.Model | None
        Model class or indicator for setting up the corresponding model class.

    Attributes
    ----------
    type
        String indicator for model.
    model : lmfit.models.Model | None
        The model definition (return value of :func:`scipy.interpolate.splrep`)
    model_result : lmfit.model.ModelResult, collection of model results.
        The results collected from fitting the model to specified data.
    """

    def __init__(
        self,
        type: (
            Literal["none", "zero", "one", "constant", "linear", "polynomial", "spline"]
            | ModelLmf
            | None
        ) = None,
        **kwargs: Any,
    ) -> None:
        self.type = type
        self.model: DriftModel | None
        self.model_result = None

        if type is None:
            self.type = "none"
            self.model = None
        elif type == "zero":
            self.model = _ConstantZeroModelFacade()
        elif type == "one":
            self.model = _ConstantOneModelFacade()
        elif type == "constant":
            self.model = _ConstantModelFacade(**kwargs)
        elif type == "linear":
            self.model = _LmfitModelFacade(LinearModel(**kwargs))
        elif type == "polynomial":
            self.model = _LmfitModelFacade(
                PolynomialModel(**dict(dict(degree=3), **kwargs))
            )
        elif getattr(type, "__module__", None) == "lmfit.models":
            self.type = type.name  # type: ignore
            self.model = _LmfitModelFacade(model=type)
        elif type == "spline":
            self.model = _SplineModelFacade(**kwargs)
        else:
            raise TypeError(f"DriftComponent cannot handle type={type}.")

    def fit(
        self,
        x: npt.ArrayLike,
        y: npt.ArrayLike,
        verbose: bool = False,
        **kwargs: Any,
    ) -> Self:
        """
        Fit model to the given data and create `self.model_results`.

        Parameters
        ----------
        x
            x data
        y
            y values
        verbose
            show plot
        kwargs
            Other parameters passed to :func:`lmfit.model.fit` or to
            :func:`scipy.interpolate.splrep`
            Use the parameter `s` to set the amount of smoothing.

        Returns
        -------
        Self
        """
        if self.model is None:
            raise ValueError("No model available.")

        self.model_result = self.model.fit(x, y, verbose=verbose, **kwargs)
        return self

    def eval(self, x: npt.ArrayLike) -> float | npt.NDArray[np.float64] | list[Any]:
        """
        Compute a transformation for time `x` from the drift model.

        Parameters
        ----------
        x
            frame values

        Returns
        -------
        npt.NDArray[np.float64].
        """
        if self.model is None:
            raise ValueError("No model available.")
        return_value: float | npt.NDArray[np.float64] | list[Any] = self.model.eval(x)
        return return_value


class Drift(_Analysis):
    """
    Estimate drift from localization coordinates by registering points in
    successive time-chunks of localization
    data using an iterative closest point algorithm (icp) or image
    cross-correlation algorithm (cc).

    Parameters
    ----------
    locdata : LocData
        Localization data representing the source on which to perform the manipulation.
    chunks : Sequence[tuple[int, ...]] | None
        Localization chunks as defined by a list of index-tuples.
        One of `chunks`, `chunk_size` or `n_chunks` must be different
        from None.
    chunk_size : int | None
        Number of consecutive localizations to form a single chunk of data.
        One of `chunks`, `chunk_size` or `n_chunks` must be different
        from None.
    n_chunks : int | None
        Number of chunks.
        One of `chunks`, `chunk_size` or `n_chunks` must be different
        from None.
    target : Literal["first", "previous"]
        The chunk on which all other chunks are aligned.
        One of 'first', 'previous'.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    method : Literal["cc", "icp"]
        The method used for computation.
        One of iterative closest point algorithm 'icp' or image
        cross-correlation algorithm 'cc'.
    bin_size : tuple
        Only for method='cc': Size per image pixel
    kwargs_chunk : dict[str, Any] | None
        Other parameter passed to :meth:`LocData.from_chunks`.
    kwargs_icp : dict[str, Any] | None
        Other parameter passed to :func:`_register_icp_open3d`.
    kwargs_cc : dict[str, Any] | None
        Other parameter passed to :func:`register_cc`.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    locdata : LocData | None
        Localization data representing the source on which to perform the manipulation.
    collection : LocData | None
        Collection of locdata chunks
    transformations : list[Transformation] | None
        Transformations for locdata chunks
    transformation_models : dict[str, list[DriftModel | DriftComponent] | None]
        The fitted model objects.
    locdata_corrected : LocData | None
        Localization data with drift-corrected coordinates.
    """

    count = 0

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        chunks: list[tuple[int, ...]] | None = None,
        chunk_size: int | None = None,
        n_chunks: int | None = None,
        target: Literal["first", "previous"] = "first",
        method: Literal["cc", "icp"] = "icp",
        kwargs_chunk: dict[str, Any] | None = None,
        kwargs_register: dict[str, Any] | None = None,
    ):
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.locdata: LocData | None = None
        self.collection: LocData | None = None
        self.transformations: list[Transformation] | None = None
        self.transformation_models: dict[
            str, list[DriftModel | DriftComponent] | None
        ] = dict(matrix=None, offset=None)
        self.locdata_corrected: LocData | None = None

    def __bool__(self) -> bool:
        if self.transformations is not None:
            return True
        else:
            return False

    def compute(self, locdata: LocData) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData
            Localization data representing the source on which to perform the
            manipulation.

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        if self.parameter["method"] == "icp":
            collection, transformations = _estimate_drift_icp(
                locdata,
                chunks=self.parameter["chunks"],
                chunk_size=self.parameter["chunk_size"],
                n_chunks=self.parameter["n_chunks"],
                kwargs_chunk=self.parameter["kwargs_chunk"],
                kwargs_register=self.parameter["kwargs_register"],
            )
        elif self.parameter["method"] == "cc":
            collection, transformations = _estimate_drift_cc(
                locdata,
                chunks=self.parameter["chunks"],
                chunk_size=self.parameter["chunk_size"],
                n_chunks=self.parameter["n_chunks"],
                kwargs_chunk=self.parameter["kwargs_chunk"],
                kwargs_register=self.parameter["kwargs_register"],
            )
        else:
            raise ValueError(
                f'Method {self.parameter["method"]} is not defined. One of "icp", "cc".'
            )
        self.locdata = locdata
        self.collection = collection
        self.transformations = transformations
        self.transformation_models = dict(matrix=None, offset=None)
        self.locdata_corrected = None
        return self

    def _transformation_models_for_identity_matrix(
        self,
    ) -> dict[str, list[DriftModel | DriftComponent]]:
        """
        Return transformation_models (dict) with DriftModels according to unit
        matrix.
        """
        dimension = self.locdata.dimension  # type: ignore[union-attr]
        transformation_models: list[DriftModel | DriftComponent] = []
        for k in np.identity(dimension).flatten():
            if k == 0:
                transformation_models.append(DriftComponent("zero"))
            else:  # if k == 1
                transformation_models.append(DriftComponent("one"))
        return dict(matrix=transformation_models)

    def _transformation_models_for_zero_offset(
        self,
    ) -> dict[str, list[DriftModel | DriftComponent]]:
        """
        Return transformation_models (dict) with DriftModels according to zero
        offset.
        """
        dimension = self.locdata.dimension  # type: ignore[union-attr]
        return dict(offset=[DriftComponent("zero") for _ in range(dimension)])

    def fit_transformation(
        self,
        slice_data: slice | None = None,
        transformation_component: Literal["matrix", "offset"] = "matrix",
        element: int = 0,
        drift_model: DriftComponent | str | None = "linear",
        verbose: bool = False,
    ) -> Self:
        """
        Fit drift model to selected component of the estimated transformations.

        Parameters
        ----------
        slice_data
            Reduce data to a selection on the localization chunks.
        transformation_component
            One of 'matrix' or 'offset'
        element
            The element of flattened transformation matrix or offset
        drift_model
            A drift model as defined by a :class:`DriftComponent` instance
            or the parameter `type` as defined in :class:`DriftComponent`.
            For None no change will occur. To reset transformation_models set
            the transformation_component to None:
            e.g. self.transformation_components = None or
            self.transformation_components['matrix'] = None.
        verbose
            Print the fitted model curves using lmfit.

        Returns
        -------
        Self
        """
        if self.transformations is None:
            logger.warning("No transformations available to be fitted.")
            return self

        if slice_data is None:
            slice_data = slice(None)

        dimension = self.locdata.dimension  # type: ignore[union-attr]
        # frames = np.array([locdata.data.frame.mean() for locdata in self.collection.references[1:]])[slice_data]
        frames = np.array(
            [locdata.data.frame.mean() for locdata in self.collection.references]  # type: ignore[union-attr]
        )[slice_data]

        if drift_model is None:
            return self

        if not isinstance(drift_model, DriftComponent):
            drift_model = DriftComponent(type=drift_model)

        if self.parameter["target"] == "first":
            if transformation_component == "matrix":
                y = np.array(
                    [
                        transformation.matrix[element // dimension][element % dimension]
                        for transformation in self.transformations
                    ]
                )[slice_data]
                if self.transformation_models["matrix"] is None:
                    self.transformation_models.update(
                        self._transformation_models_for_identity_matrix()
                    )
            elif transformation_component == "offset":
                y = np.array(
                    [
                        transformation.offset[element]
                        for transformation in self.transformations
                    ]
                )[slice_data]
                if self.transformation_models["offset"] is None:
                    self.transformation_models.update(
                        self._transformation_models_for_zero_offset()
                    )
            else:
                raise ValueError(
                    "transformation_component must be 'matrix' or 'offset'."
                )

            self.transformation_models[transformation_component][element] = drift_model  # type: ignore[index]
            self.transformation_models[transformation_component][element].fit(  # type: ignore[index]
                frames, y, verbose=False
            )

            if verbose:
                self.transformation_models[transformation_component][  # type: ignore[index,union-attr]
                    element
                ].model.plot()

        elif self.parameter["target"] == "previous":
            raise NotImplementedError("Not yet implemented. Use target = 'first'.")

        return self

    def fit_transformations(
        self,
        slice_data: slice | None = None,
        matrix_models: list[DriftComponent] | None = None,
        offset_models: list[DriftComponent] | None = None,
        verbose: bool = False,
    ) -> Self:
        """
        Fit model parameter to all estimated transformation components.

        Parameters
        ----------
        slice_data
            Reduce data to a selection on the localization chunks.
        matrix_models
            Models to use for fitting each matrix component.
            Length of list must be equal to the square of the transformations
            dimension (4 or 9).
            If None, no matrix transformation will be carried out when calling
            :func:`apply_correction`.
        offset_models
            Models to use for fitting each offset component.
            Length of list must be equal to the transformations dimension
            (2 or 3).
            If None, no offset transformation will be carried out when calling
            :func:`apply_correction`.
        verbose
            Print the fitted model curves.

        Returns
        -------
        Self
        """
        if self.transformations is None:
            logger.warning("No transformations available to be fitted.")
            return self

        if slice_data is None:
            slice_data = slice(None)

        if matrix_models is None:
            self.transformation_models["matrix"] = None
        else:
            if len(matrix_models) != self.locdata.dimension**2:  # type: ignore[union-attr]
                raise ValueError(
                    "Length of matrix_models must be equal to the square of the "
                    "transformations dimension (4 or 9)."
                )
            self.transformation_models.update(
                self._transformation_models_for_identity_matrix()
            )
            for n, matrix_model in enumerate(matrix_models):
                if matrix_model is None:
                    matrix_model = self.transformation_models["matrix"][n]
                self.fit_transformation(
                    slice_data=slice_data,
                    transformation_component="matrix",
                    element=n,
                    drift_model=matrix_model,
                    verbose=verbose,
                )

        if offset_models is None:
            self.transformation_models["offset"] = None
        else:
            if len(offset_models) != self.locdata.dimension:  # type: ignore[union-attr]
                raise ValueError(
                    "Length of offset_models must be equal to the the transformations dimension (2 or 3)."
                )
            self.transformation_models.update(
                self._transformation_models_for_zero_offset()
            )
            for n, offset_model in enumerate(offset_models):
                if offset_model is None:
                    offset_model = self.transformation_models["offset"][n]
                self.fit_transformation(
                    slice_data=slice_data,
                    transformation_component="offset",
                    element=n,
                    drift_model=offset_model,
                    verbose=verbose,
                )

        return self

    def _apply_correction_on_chunks(self) -> LocData:
        """
        Correct drift by applying the estimated transformations to locdata chunks.
        """
        transformed_locdatas: list[LocData] = []
        assert self.transformations is not None  # type narrowing # noqa: S101
        assert (  # type narrowing # noqa: S101
            self.collection is not None
            and isinstance(self.collection.references, Sequence)
        )

        if self.parameter["target"] == "first":
            transformed_locdatas = [
                transform_affine(locdata, transformation.matrix, transformation.offset)  # type: ignore
                for locdata, transformation in zip(
                    self.collection.references[1:], self.transformations
                )
            ]

        elif self.parameter["target"] == "previous":
            for n, locdata in enumerate(self.collection.references[1:]):
                transformed_locdata = locdata
                for transformation in reversed(self.transformations[:n]):
                    transformed_locdata = transform_affine(
                        transformed_locdata,
                        transformation.matrix,
                        transformation.offset,
                    )
                transformed_locdatas.append(transformed_locdata)

        new_locdata = LocData.concat(
            [self.collection.references[0]] + transformed_locdatas
        )
        return new_locdata

    def _apply_correction_from_model(self, locdata: LocData) -> npt.NDArray[Any]:
        """
        Correct drift by applying the estimated transformations to locdata.
        If self.transformation_model['matrix'] is None, no matrix
        transformation will be carried out when calling
        :func:`apply_correction` (same for 'offset').

        Parameters
        ----------
        locdata
            Localization data to apply correction on. If None correction is
            applied to self.locdata.
        """
        # check if any models are fitted,
        # otherwise it is likely that the fitting procedure was accidentally omitted.
        if (
            self.transformation_models["matrix"] is None
            and self.transformation_models["offset"] is None
        ):
            raise AttributeError(
                "The transformation_models have to be fitted before they can be evaluated."
            )

        dimension = locdata.dimension
        frames = locdata.data.frame.values
        matrix = np.tile(np.identity(dimension), (len(frames), *([1] * dimension)))
        offset = np.zeros((len(frames), dimension))

        if self.transformation_models["matrix"] is None:
            transformed_points = locdata.coordinates
        else:
            for n, drift_model in enumerate(self.transformation_models["matrix"]):
                matrix[:, n // dimension, n % dimension] = drift_model.eval(frames)  # type: ignore
            transformed_points = np.einsum(
                "...ij, ...j -> ...i", matrix, locdata.coordinates
            )

        if self.transformation_models["offset"] is None:
            pass
        else:
            for n, drift_model in enumerate(self.transformation_models["offset"]):
                offset[:, n] = drift_model.eval(frames)  # type: ignore
            transformed_points += offset

        return transformed_points

    def apply_correction(
        self, locdata: LocData | None = None, from_model: bool = True
    ) -> Self:
        """
        Correct drift by applying the estimated transformations to locdata.

        Parameters
        ----------
        locdata
            Localization data to apply correction on. If None correction is
            applied to self.locdata.
        from_model
            If `True` compute transformation matrix from fitted transformation
            models and apply interpolated
            transformations. If False use the estimated transformation matrix
            for each data chunk.

        Returns
        -------
        Self
        """
        if self.transformations is None:
            logger.warning("No transformations available to be applied.")
            return self

        local_parameter = locals()

        if locdata is None:
            locdata_orig: LocData = self.locdata  # type: ignore[assignment]
        else:
            locdata_orig = locdata

        if locdata_orig is None:
            logger.warning("Locdata is None.")
            self.locdata_corrected = locdata_orig
            return self

        elif not len(locdata_orig):
            logger.warning("Locdata is empty.")
            self.locdata_corrected = locdata_orig
            return self

        if from_model:
            transformed_points = self._apply_correction_from_model(locdata=locdata_orig)
        else:
            if locdata is not None:
                raise TypeError(
                    "Locdata must be None since correction can only be applied to original locdata chunks."
                )
            transformed_points = self._apply_correction_on_chunks().coordinates

        # new LocData object
        new_dataframe = locdata_orig.data.copy()
        df = pd.DataFrame(
            transformed_points,
            columns=locdata_orig.coordinate_keys,
            index=locdata_orig.data.index,
        )
        # cast dtypes
        df = df.astype(new_dataframe[locdata_orig.coordinate_keys].dtypes)
        new_dataframe.update(df)
        new_locdata = LocData.from_dataframe(new_dataframe)

        # update metadata
        meta_ = _modify_meta(
            locdata_orig,
            new_locdata,
            function_name=sys._getframe().f_code.co_name,
            parameter=local_parameter,
            meta=None,
        )
        new_locdata.meta = meta_

        self.locdata_corrected = new_locdata
        return self

    def plot(
        self,
        ax: mpl.axes.Axes | None = None,
        transformation_component: Literal["matrix", "offset"] = "matrix",
        element: int | None = None,
        window: int = 1,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Plot the transformation components as function of average frame for
        each locdata chunk.

        Parameters
        ----------
        ax
            The axes on which to show the image
        transformation_component
            One of 'matrix' or 'offset'
        element
            The element of flattened transformation matrix or offset to be
            plotted; if None all plots are shown.
        window
            Window for running average that is applied before plotting.
            Not implemented yet.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.transformations is None:
            return ax

        n_transformations = len(self.transformations)
        # prepare plot
        x = [reference.data.frame.mean() for reference in self.collection.references]  # type: ignore[union-attr]
        results = np.array(
            [
                getattr(transformation, transformation_component)
                for transformation in self.transformations
            ]
        )
        if element is None:
            ys = results.reshape(n_transformations, -1).T
            for i, y in enumerate(ys):
                ax.plot(
                    x,
                    y,
                    **dict(dict(label=f"{transformation_component}[{i}]"), **kwargs),
                )
        else:
            y = results.reshape(n_transformations, -1).T[element]
            ax.plot(
                x,
                y,
                **dict(dict(label=f"{transformation_component}[{element}]"), **kwargs),
            )

        ax.set(
            title=f"Drift\n (window={window})",
            xlabel="frame",
            ylabel="".join([transformation_component]),
        )

        return ax
